# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-08-09 23:21
from __future__ import unicode_literals

import django.db.models.deletion
import django_dsl.fields
import flexible_reports.models.report
import flexible_reports.models.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='Column',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('label', models.TextField(verbose_name='Label')),
                ('position', models.PositiveIntegerField(default=0, verbose_name='Position')),
                ('sortable', models.BooleanField(default=True)),
                ('attr_name', models.CharField(blank=True, help_text='\n        Attribute name on the parent table\'s base model. \n        \n        If this column is sortable, that attribute is used to sort this \n        column. \n        \n        In case no value is given in "Template" field, \n        this attribute will be used to get information from model \n        instances.\n         \n        Warning, if you want to make this column sortable, you need to \n        provide this value.\n        \n        This value can contain dot notation to reference fields in related \n        models.\n        ', max_length=200, null=True, verbose_name='Attribute name')),
                ('template', models.TextField(blank=True, default='{{ value }}', help_text='If empty, the value of the object\'s attribute from \n        "Attribute name" field will be used instead. \n            \n        Template will get following values in it\'s context:     \n        - *record*  -- data record for the current row\n        - *value*   -- value from `record` that corresponds to the current column\n        - *default* -- appropriate default value to use as fallback\n        ', null=True, verbose_name='Template')),
                ('display_totals', models.BooleanField(default=False, help_text='Display column totals in footer. For columns without Attribute name, this will be total number of the rows.', verbose_name='Display totals')),
                ('footer_template', models.TextField(blank=True, default='{{ value }}', help_text='\n        Template for footer.', null=True, verbose_name='Footer template')),
            ],
            options={
                'verbose_name': 'Column',
                'verbose_name_plural': 'Columns',
                'ordering': ('position',),
            },
        ),
        migrations.CreateModel(
            name='Datasource',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('label', models.TextField(verbose_name='Label')),
                ('dsl_query', django_dsl.fields.DjangoDSLField(verbose_name='DSL query')),
                ('distinct', models.BooleanField(default=True, help_text='Output only distinct records', verbose_name='Distinct')),
                ('base_model', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType', verbose_name='Base model')),
            ],
            options={
                'verbose_name': 'Datasource',
                'verbose_name_plural': 'Datasources',
                'ordering': ('label',),
            },
        ),
        migrations.CreateModel(
            name='Report',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('title', models.TextField(verbose_name='Title')),
                ('subtitle', models.TextField(blank=True, null=True, verbose_name='Subtitle')),
                ('slug', models.SlugField()),
                ('template', models.TextField(default=flexible_reports.models.report.get_reports_template, validators=[flexible_reports.models.validators.TemplateValidator], verbose_name='Template')),
            ],
            options={
                'verbose_name': 'Report',
                'verbose_name_plural': 'Reports',
            },
        ),
        migrations.CreateModel(
            name='ReportElement',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('title', models.TextField(verbose_name='Title')),
                ('subtitle', models.TextField(blank=True, null=True, verbose_name='Subtitle')),
                ('position', models.PositiveIntegerField(default=0, verbose_name='Position')),
                ('slug', models.SlugField(max_length=200)),
                ('datasource', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='flexible_reports.Datasource')),
                ('parent', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='flexible_reports.Report')),
            ],
            options={
                'ordering': ('position',),
            },
        ),
        migrations.CreateModel(
            name='Table',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('label', models.TextField(verbose_name='Label')),
                ('sort_option', models.IntegerField(choices=[(0, "sort with other tables - if a column in this table is sorted, all the other tables\n    on the page with columns with identical name will be sorted too, as long \n    as they also are marked as 'sort with other tables'\n    "), (1, 'sort individually - this table will be sorted individually; even if other \n    tables on the page have columns with same label this table will be \n    sorted independently'), (2, 'sort in group - this table will be sorted together with a group of \n    tables; you must enter a prefix for this group')], default=0, verbose_name='Sort option')),
                ('group_prefix', models.CharField(blank=True, help_text='this value is used as a prefix only when "Sort \n        option" is set to "sort in group"\n        ', max_length=200, null=True, verbose_name='Group prefix')),
                ('empty_template', models.TextField(blank=True, default='There is no data for this table.', help_text='\n        Template which will be displayed when there is no data for this \n        table.\n        ', null=True, verbose_name='Empty template')),
                ('base_model', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType', verbose_name='Base model')),
            ],
            options={
                'verbose_name': 'Table',
                'verbose_name_plural': 'Tables',
            },
        ),
        migrations.AddField(
            model_name='reportelement',
            name='table',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='flexible_reports.Table'),
        ),
        migrations.AddField(
            model_name='column',
            name='parent',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='flexible_reports.Table'),
        ),
        migrations.AlterUniqueTogether(
            name='reportelement',
            unique_together=set([('parent', 'slug'), ('parent', 'position')]),
        ),
        migrations.AlterUniqueTogether(
            name='column',
            unique_together=set([('parent', 'id', 'position')]),
        ),
    ]



# -*- encoding: utf-8 -*-
from django import forms
from django.contrib import admin
from django.utils.translation import string_concat, ugettext_lazy as _
from flexible_reports.models.table import AllSortOptions, SortInGroup

from ..models import Column, ColumnOrder, Table
from .helpers import AverageTextarea, SmallerTextarea, SortableHiddenMixin


class ColumnForm(forms.ModelForm):
    class Meta:
        widgets = {
            'label': SmallerTextarea,
            'template': AverageTextarea,
            'footer_template': SmallerTextarea,
            'attrs': SmallerTextarea
        }


class ColumnOrderForm(forms.ModelForm):
    def __init__(self, parent, *args, **kw):
        super(ColumnOrderForm, self).__init__(*args, **kw)
        self.fields['column'].queryset = Column.objects.filter(parent=parent)


class ColumnOrderInline(SortableHiddenMixin, admin.TabularInline):
    extra = 0
    model = ColumnOrder
    fields = ['column', 'desc', 'position']

    def formfield_for_foreignkey(self, db_field, request=None, **kwargs):
        field = super(ColumnOrderInline, self).formfield_for_foreignkey(db_field, request, **kwargs)
        if db_field.name == 'column':
            if request._parent_ is not None:
                field.queryset = field.queryset.filter(
                    parent=request._parent_,
                    sortable=True)

            else:
                field.queryset = field.queryset.none()
        return field


class ColumnInline(SortableHiddenMixin, admin.StackedInline):
    extra = 0
    model = Column
    form = ColumnForm
    fields = ['label',
              'attr_name',
              'template',
              'attrs',
              'sortable',
              'exclude_from_export',
              'strip_html_on_export',
              'display_totals',
              'footer_template',
              'position']


class TableForm(forms.ModelForm):
    class Meta:
        fields = ['label',
                  'base_model',
                  'sort_option',
                  'group_prefix',
                  'attrs',
                  'empty_template',
                  ]
        widgets = {
            'label': SmallerTextarea,
            'empty_template': SmallerTextarea,
            'attrs': SmallerTextarea
        }

    pass


@admin.register(Table)
class TableAdmin(admin.ModelAdmin):
    list_display = ['label',
                    'base_model',
                    'short_sort_option',
                    'columns']
    inlines = [ColumnInline, ColumnOrderInline]
    form = TableForm

    def columns(self, obj):
        return ", ".join([x.label for x in obj.column_set.all()])

    columns.short_description = _("Columns")

    def short_sort_option(self, obj):
        if obj.sort_option == SortInGroup.id:
            return string_concat(
                SortInGroup.label,
                _(" (group name: "),
                obj.group_prefix,
                ")"
            )
        return AllSortOptions[obj.sort_option].label

    short_sort_option.short_description = _("Sort option")
    short_sort_option.admin_order_field = "sort_option"

    def get_form(self, request, obj=None, **kwargs):
        request._parent_ = obj
        return super(TableAdmin, self).get_form(request, obj, **kwargs)



# -*- encoding: utf-8 -*-
from django.forms.widgets import Textarea

SmallerTextarea = Textarea(attrs={'cols': 75, 'rows': 2})
AverageTextarea = Textarea(attrs={'cols': 75, 'rows': 4})
BiggerTextarea = Textarea(attrs={'cols': 75, 'rows': 18})

try:
    from grappelli.forms import GrappelliSortableHiddenMixin as \
        SortableHiddenMixin
except ImportError:
    class SortableHiddenMixin:
        # no-op
        pass



# -*- encoding: utf-8 -*-

from django import forms
from django.contrib import admin
from django.core.exceptions import ValidationError
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext_lazy as _
from django_dsl import compiler, exceptions
from flexible_reports.admin.helpers import BiggerTextarea

from .. import utils
from ..models.datasource import Datasource
from .helpers import SmallerTextarea


class DatasourceForm(forms.ModelForm):
    def clean(self):
        if 'base_model' not in self.cleaned_data:
            return

        model = self.cleaned_data['base_model'].model_class()
        shortcuts = utils.get_shortcuts(model)

        try:
            filter = compiler.compile(
                self.cleaned_data['dsl_query'],
                shortcuts)
        except exceptions.CompileException:
            return  # handled by dsl_query.validators

        try:
            model.objects.filter(filter).first()
        except Exception as e:
            raise ValidationError(
                {"dsl_query": [
                    ValidationError(
                        _("An error occured while trying to run the actual "
                          "database query (%(error)s)"),
                        params={"error": e}
                    )
                ]}
            )

    class Meta:
        model = Datasource
        fields = ["label", "base_model", "dsl_query", "distinct"]
        widgets = {
            'label': SmallerTextarea,
            'dsl_query': BiggerTextarea
        }


@admin.register(Datasource)
class DatasourceAdmin(admin.ModelAdmin):
    list_display = ['label', 'base_model', 'dsl_query_fmt']
    form = DatasourceForm

    def dsl_query_fmt(self, obj):
        return mark_safe(f"<pre>{ obj.dsl_query }</pre>")

    dsl_query_fmt.short_description = _("DSL query")



# -*- encoding: utf-8 -*-
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.db import models
from django.template.loader import get_template
from django.utils.translation import ugettext_lazy as _

from .behaviors import Orderable, Titled
from .validators import TemplateValidator

DATA_FROM_DATASOURCE = 0
DATA_FROM_EXCEPT_CATCHALL = 2


class ReportElement(Titled, Orderable):
    parent = models.ForeignKey('flexible_reports.Report',
                               verbose_name=_("Parent"),
                               on_delete=models.CASCADE
                               )
    data_from = models.IntegerField(
        verbose_name=_("Data from"),
        choices=[
            (DATA_FROM_DATASOURCE, _("datasource")),
            (DATA_FROM_EXCEPT_CATCHALL, _("except catchall"))
        ],
        default=DATA_FROM_DATASOURCE)

    datasource = models.ForeignKey(
        'flexible_reports.Datasource',
        verbose_name=_("Datasource"),
        null=True,
        blank=True,
        help_text=_("Fill only if 'datasource' has been chosen in 'Data from' "
                    "field"),
        on_delete=models.CASCADE
    )

    base_model = models.ForeignKey(
        ContentType,
        verbose_name=_("Base model"),
        null=True, blank=True,
        help_text=_(
            "Fill only if 'except catchall' is selected in 'Data from' "
            "field"),
        on_delete=models.CASCADE
    )

    table = models.ForeignKey('flexible_reports.Table',
                              verbose_name=_("Table"),
                              on_delete=models.CASCADE)
    slug = models.SlugField(max_length=200, verbose_name=_("Slug"))

    def clean(self):
        if self.data_from != DATA_FROM_DATASOURCE:
            if self.datasource is not None:
                raise ValidationError(
                    {"datasource": [ValidationError(
                        _(
                            "In case when data is from except-catchall, "
                            "please specify an empty datasource."))]})

            if self.base_model is None:
                raise ValidationError(
                    {"base_model": [ValidationError(
                        _(
                            "In case when data is from except-catchall, "
                            "please specify a base model."))]})

        if self.data_from == DATA_FROM_DATASOURCE:
            if self.datasource is None:
                raise ValidationError(
                    {"datasource": [ValidationError(
                        _("Please specify a datasource."))]})

            if self.base_model is not None:
                raise ValidationError(
                    {"base_model": [ValidationError(
                        _("Please specify an empty base model."))]})

    class Meta:
        unique_together = [
            ('parent', 'slug')
        ]
        ordering = ('position',)
        verbose_name = _("Report element")
        verbose_name_plural = _("Report elements")


def _get_template(name):
    t = get_template(name)
    return t.template.source


def get_reports_template():
    return _get_template("flexible_reports/report.html")


class Report(Titled):
    """Report is a collection of multiple Datasources' data
    rendered in Tables. """

    slug = models.SlugField()

    template = models.TextField(
        verbose_name=_("Template"),
        default=get_reports_template,
        validators=[TemplateValidator, ]
    )

    _base_queryset = None
    _context = None

    @property
    def base_queryset(self):
        if self._base_queryset is None:
            raise ImproperlyConfigured("Please set base queryset for this "
                                       "report using Report.set_base_queryset")
        return self._base_queryset

    @property
    def context(self):
        return self._context

    class Meta:
        verbose_name_plural = _("Reports")
        verbose_name = _("Report")

    def set_base_queryset(self, queryset):
        self._base_queryset = queryset

    def set_context(self, context):
        self._context = context



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-10-26 08:40
from __future__ import unicode_literals

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('flexible_reports', '0009_auto_20171025_0558'),
    ]

    operations = [
        migrations.AddField(
            model_name='reportelement',
            name='base_model',
            field=models.ForeignKey(blank=True, help_text="Fill only if 'except catchall' is selected in 'Data from' field", null=True, on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType', verbose_name='Base model'),
        ),
        migrations.AlterField(
            model_name='reportelement',
            name='data_from',
            field=models.IntegerField(choices=[(0, 'datasource'), (2, 'except catchall')], default=0, verbose_name='Data from'),
        ),
        migrations.AlterField(
            model_name='reportelement',
            name='datasource',
            field=models.ForeignKey(blank=True, help_text="Fill only if 'datasource' has been chosen in 'Data from' field", null=True, on_delete=django.db.models.deletion.CASCADE, to='flexible_reports.Datasource', verbose_name='Datasource'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-08-23 23:42
from __future__ import unicode_literals

import django.contrib.postgres.fields.jsonb
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('flexible_reports', '0003_table_attrs'),
    ]

    operations = [
        migrations.AlterField(
            model_name='table',
            name='attrs',
            field=django.contrib.postgres.fields.jsonb.JSONField(blank=True, null=True, verbose_name='HTML attributes'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-10-25 10:53
from __future__ import unicode_literals

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('flexible_reports', '0007_sort_desc'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='columnorder',
            options={'ordering': ('position',), 'verbose_name': 'Column order information', 'verbose_name_plural': 'Column order informations'},
        ),
        migrations.AddField(
            model_name='reportelement',
            name='data_from',
            field=models.IntegerField(choices=[(0, 'datasource'), (1, 'catchall'), (2, 'except catchall')], default=0, verbose_name='Data from'),
        ),
        migrations.AlterField(
            model_name='reportelement',
            name='datasource',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='flexible_reports.Datasource', verbose_name='Datasource'),
        ),
        migrations.AlterField(
            model_name='reportelement',
            name='parent',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='flexible_reports.Report', verbose_name='Parent'),
        ),
        migrations.AlterField(
            model_name='reportelement',
            name='slug',
            field=models.SlugField(max_length=200, verbose_name='Slug'),
        ),
        migrations.AlterField(
            model_name='reportelement',
            name='table',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='flexible_reports.Table', verbose_name='Table'),
        ),
        migrations.AlterUniqueTogether(
            name='reportelement',
            unique_together=set([('parent', 'slug')]),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-08-30 23:11
from __future__ import unicode_literals

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('flexible_reports', '0005_column_attrs'),
    ]

    operations = [
        migrations.CreateModel(
            name='ColumnOrder',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('position', models.PositiveIntegerField(default=0, verbose_name='Position')),
                ('column', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='flexible_reports.Column', verbose_name='Column')),
                ('table', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='flexible_reports.Table', verbose_name='Table')),
            ],
            options={
                'ordering': ('position',),
                'abstract': False,
            },
        ),
    ]



# -*- coding: utf-8 -*-

from collections import OrderedDict

from django.contrib.contenttypes.models import ContentType
from django.contrib.postgres.fields.jsonb import JSONField
from django.core.exceptions import ValidationError
from django.db import models
from django.utils.translation import string_concat, ugettext_lazy as _

from .behaviors import Labelled, Orderable


class SortWithOtherTables:
    """prefix = None"""
    id = 0
    label = _("sort with other tables")
    help_text = _("""if a column in this table is sorted, all the other tables
    on the page with columns with identical name will be sorted too, as long
    as they also are marked as 'sort with other tables'
    """)

    def get_prefix(self, table):
        return ""


class SortIndividually:
    """prefix = self.pk"""

    id = 1
    label = _("sort individually")
    help_text = _("""this table will be sorted individually; even if other
    tables on the page have columns with same label this table will be
    sorted independently""")

    def get_prefix(self, table):
        return table.pk


class SortInGroup:
    """prefix = self.group_prefix"""
    id = 2
    label = _("sort in group")
    help_text = _("""this table will be sorted together with a group of
    tables; you must enter a prefix for this group""")

    def get_prefix(self, table):
        return table.group_prefix


AllSortOptions = OrderedDict(
    [(x.id, x) for x in [SortWithOtherTables,
                         SortIndividually,
                         SortInGroup]])


class ColumnOrder(Orderable):
    table = models.ForeignKey("flexible_reports.Table",
                              verbose_name=_("Table"),
                              on_delete=models.CASCADE)
    column = models.ForeignKey("flexible_reports.Column",
                               verbose_name=_("Column"),
                               on_delete=models.CASCADE)
    desc = models.BooleanField(
        _("Descending"),
        default=False
    )

    class Meta:
        verbose_name = _("Column order information")
        verbose_name_plural = _("Column order informations")
        ordering = ('position',)

    def get(self):
        if not self.desc:
            return self.column.label
        return f"-{ self.column.label }"


class Table(Labelled):
    """Collection of Columns."""

    base_model = models.ForeignKey(
        ContentType,
        verbose_name=_("Base model"),
        on_delete=models.CASCADE)

    sort_option = models.IntegerField(
        default=0,
        verbose_name=_("Sort option"),
        choices=[
            (x.id, string_concat(x.label, " - ", x.help_text))
            for x in AllSortOptions.values()
        ],
    )

    attrs = JSONField(
        verbose_name=_("HTML attributes"),
        blank=True,
        null=True
    )

    group_prefix = models.CharField(
        verbose_name=_("Group prefix"),
        null=True,
        blank=True,
        max_length=200,
        help_text=_("""this value is used as a prefix only when "Sort
        option" is set to "sort in group"
        """))

    empty_template = models.TextField(
        verbose_name=_("Empty template"),
        null=True,
        blank=True,
        help_text=_("""
        Template which will be displayed when there is no data for this
        table.
        """),
        default=_("There is no data for this table.")
    )

    class Meta:
        verbose_name = _("Table")
        verbose_name_plural = _("Tables")

    def clean(self):
        if self.sort_option == SortInGroup.id:
            if not self.group_prefix:
                raise ValidationError({
                    "group_prefix": [ValidationError(
                        _("Please enter group prefix if you want to sort in "
                          "group")
                    )]
                })

    def get_prefix(self):
        return AllSortOptions[self.sort_option].get_prefix(None, self)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import re
import sys
from distutils.cmd import Command
from distutils.command.build import build as _build

from setuptools.command.install_lib import install_lib as _install_lib

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


def get_version(*file_paths):
    """Retrieves the version from flexible_reports/__init__.py"""
    filename = os.path.join(os.path.dirname(__file__), *file_paths)
    version_file = open(filename).read()
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError('Unable to find version string.')


version = get_version("flexible_reports", "__init__.py")


class compile_translations(Command):
    description = 'compile message catalogs to MO files via django compilemessages'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        curdir = os.getcwd()
        os.chdir(os.path.realpath('flexible_reports'))
        from django.core.management import call_command
        call_command('compilemessages')
        os.chdir(curdir)


class build(_build):
    sub_commands = [('compile_translations', None)] + _build.sub_commands


class install_lib(_install_lib):
    def run(self):
        self.run_command('compile_translations')
        _install_lib.run(self)


if sys.argv[-1] == 'publish':
    try:
        import wheel

        print("Wheel version: ", wheel.__version__)
    except ImportError:
        print('Wheel library missing. Please run "pip install wheel"')
        sys.exit()
    os.system('python setup.py sdist upload')
    os.system('python setup.py bdist_wheel upload')
    sys.exit()

if sys.argv[-1] == 'tag':
    print("Tagging the version on git:")
    os.system("git tag -a %s -m 'version %s'" % (version, version))
    os.system("git push --tags")
    sys.exit()

readme = open('README.rst').read()
history = open('HISTORY.rst').read().replace('.. :changelog:', '')

setup(
    name='django-flexible-reports',
    version=version,
    description="""A framework for report generation in Django""",
    long_description=readme + '\n\n' + history,
    author='Michał Pasternak',
    author_email='michal.dtz@gmail.com',
    url='https://github.com/mpasternak/django-flexible-reports',
    packages=[
        'flexible_reports',
    ],
    include_package_data=True,
    install_requires=['Django', ] + [x.strip() for x in open(
        os.path.join(os.path.dirname(__file__),
                     "requirements.txt")).readlines() if x.strip()],
    license="MIT",
    zip_safe=False,
    keywords='django-flexible-reports',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Framework :: Django',
        'Framework :: Django :: 1.10',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.6',
    ],
    cmdclass={'build': build, 'install_lib': install_lib,
              'compile_translations': compile_translations}
)



# -*- encoding: utf-8 -*-
import copy
import itertools
import sys
from collections import OrderedDict, defaultdict
from tempfile import NamedTemporaryFile

import bleach
import lxml.html
import pypandoc
from django.template.base import Context, Template
from django.utils.safestring import mark_safe
from django_tables2.columns.templatecolumn import Column, TemplateColumn
from django_tables2.export.export import TableExport
from django_tables2.tables import Table
from flexible_reports.models.report import DATA_FROM_DATASOURCE
from tablib.core import Databook, Dataset


class CounterMixin:
    def __init__(self):
        self._counter = itertools.count(1)

    def counter(self):
        return str(self._counter.__next__())


class FooterMixin:
    def __init__(self, display_totals, footer_template, kwargs):
        self.display_totals = display_totals
        self.footer_template = footer_template

        if self.display_totals:
            self.render_footer = self._render_footer

    def _render_footer(self, table):
        error = None
        try:
            value = sum([getattr(x, self.accessor)
                         for x in table.data])
        except Exception as e:
            error = str(e)
            value = len(table.data)

        context = Context({'value': value,
                           'error': error,
                           'count': len(table.data)})

        return Template(
            self.footer_template
        ).render(context=context)


class StripHTMLOnExportMixin:
    def __init__(self, strip_html_on_export):
        self.strip_html_on_export = strip_html_on_export

    def value(self, **kwargs):
        value = super(StripHTMLOnExportMixin, self).value(**kwargs)
        if self.strip_html_on_export is True:
            value = lxml.html.fromstring(str(value)).text_content()
        return value


class DjangoTables2TemplateColumn(StripHTMLOnExportMixin, CounterMixin,
                                  FooterMixin, TemplateColumn):
    def __init__(self, display_totals, footer_template, strip_html_on_export,
                 *args, **kw):
        FooterMixin.__init__(self, display_totals, footer_template, kw)
        StripHTMLOnExportMixin.__init__(self, strip_html_on_export)
        CounterMixin.__init__(self)
        TemplateColumn.__init__(self, *args, **kw)


class DjangoTables2Column(StripHTMLOnExportMixin, FooterMixin, Column):
    def __init__(self, display_totals, footer_template, strip_html_on_export,
                 *args, **kw):
        FooterMixin.__init__(self, display_totals, footer_template, kw)
        StripHTMLOnExportMixin.__init__(self, strip_html_on_export)
        Column.__init__(self, *args, **kw)


def column(column):
    kwargs = dict(
        verbose_name=column.label,
        orderable=column.sortable,
        order_by=column.attr_name,
        display_totals=column.display_totals,
        footer_template=column.footer_template,
        accessor=column.attr_name,
        exclude_from_export=column.exclude_from_export,
        strip_html_on_export=column.strip_html_on_export,
        attrs=column.attrs
    )

    if column.template:
        klass = DjangoTables2TemplateColumn(
            template_code=column.template,
            **kwargs)

    else:
        klass = DjangoTables2Column(**kwargs)

    return (column.label, klass)


_table_cache = {}


def _table(table):
    global _table_cache

    if table.pk not in _table_cache:
        order_by = []
        for column_order in (
            table.columnorder_set.all()
                .select_related()
                .order_by("position")):
            order_by.append(column_order.get())
        table.order_by = order_by

        class AdHocTable(Table):
            class Meta:
                attrs = table.attrs or {}
                order_by = table.order_by
                per_page = sys.maxsize
                prefix = table.get_prefix()
                empty_text = mark_safe(table.empty_template)

        for c in table.column_set.all():
            label, klass = column(c)
            AdHocTable.base_columns[label] = klass

        _table_cache[table.pk] = AdHocTable

    return _table_cache[table.pk]


def table(table, request, object_list):
    return _table(table)(data=object_list, request=request)


def _report(report, parent_context):
    render_context = copy.copy(parent_context)

    # elements -> an OrderedDict of
    #   table,
    #   title,
    #   object_list
    # keys are slugs of elements.
    #
    # catchall -> a dict, one entry per datasource.model,
    #   every item is a list of filters generated by datasource.get_filter
    #   it is used to catch all the items in the report, per model
    #
    # except_catchall -> a dict, one per datasource.model, contains
    #   QuerySets with all the items, that were not caught by catchall
    #
    # this may be used to render a few datasources in tables, then to
    # have a table containing everything except those records
    #

    render_context.update({
        'self': report,
        'elements': OrderedDict(),
        'catchall': defaultdict(lambda: []),
        'except_catchall': defaultdict(lambda: [])
    })

    reportelements_set = report.reportelement_set.all().prefetch_related(
        "datasource", "datasource__base_model", "table").select_related()
    for elem in reportelements_set:

        if elem.data_from == DATA_FROM_DATASOURCE:
            datasource = elem.datasource
            filter = datasource.get_filter(context=report.context)
            object_list = report.base_queryset.filter(filter)

            ds_key = "%s_%s" % (datasource.base_model.app_label,
                                datasource.base_model.model)

            render_context['catchall'][ds_key].append(filter)

            if datasource.distinct:
                object_list = object_list.distinct()

            table_dict = {
                'title': elem.title,
                'object_list': object_list,
                'table': table(
                    elem.table,
                    parent_context['request'],
                    object_list)
            }

        else:
            table_dict = {
                'except_catchall': elem.base_model,
                'title': elem.title,
                'table': elem.table
            }

        render_context['elements'][elem.slug] = table_dict

    # Fill except-catchall
    except_catchall = report.base_queryset.all()
    q = None
    for key, filters in render_context['catchall'].items():
        if not filters:
            continue

        if q is None:
            q = filters[0]
        for filter in filters[1:]:
            q |= filter

        except_catchall = except_catchall.exclude(
            pk__in=report.base_queryset.filter(q).values_list("pk", flat=True))

        render_context['except_catchall'][key] = except_catchall

    for key, elem in render_context['elements'].items():
        if 'except_catchall' not in elem:
            continue

        base_model = elem['except_catchall']

        key = "%s_%s" % (base_model.app_label, base_model.name)

        object_list = render_context['except_catchall'][key]
        elem['object_list'] = object_list
        elem['table'] = table(
            elem['table'],
            parent_context['request'],
            object_list)

    return render_context


def as_html(report, parent_context):
    render_context = _report(report, parent_context)
    return Template(report.template).render(render_context)


def as_docx(report,
            parent_context,
            allowed_tags=['table', 'tr', 'td', 'th', 'b', 'i', 'u',
                          'sup', 'sub', 'h1', 'h2', 'h3', 'h4',
                          'em', 'strong', 'strike', 'font'],
            allowed_attributes={"td": ["colspan"]}):
    data = as_html(report, parent_context)
    # Remove "<a>" tags from headers
    data = bleach.clean(data, allowed_tags, allowed_attributes, strip=True)
    f = NamedTemporaryFile(delete=False)
    pypandoc.convert_text(data, 'docx', format='html', outputfile=f.name)
    return f


def as_tablib_databook(report, parent_context):
    render_context = _report(report, parent_context)

    databook = Databook()
    for element in render_context['elements'].values():
        dataset = TableExport(
            TableExport.XLSX,
            element['table']
        ).dataset

        dataset.title = element['title'][:31]
        databook.add_sheet(dataset)

    return databook


def as_tablib_dataset(report, parent_context):
    render_context = _report(report, parent_context)

    dataset = Dataset()
    for element in render_context['elements'].values():
        table = element['table']
        dataset.append_separator(element['title'])
        for i, row in enumerate(table.as_values()):
            dataset.append(row)

    return dataset



# -*- coding: utf-8 -*-
from django.contrib.postgres.fields.jsonb import JSONField
from django.core.exceptions import ValidationError
from django.db import models
from django.utils.translation import ugettext_lazy as _

from .behaviors import Labelled, Orderable


class Column(Labelled, Orderable):
    parent = models.ForeignKey(
        'flexible_reports.Table',
        on_delete=models.CASCADE
    )

    sortable = models.BooleanField(
        verbose_name=_("Sortable"),
        default=True)

    attr_name = models.CharField(
        verbose_name=_("Attribute name"),
        max_length=200,
        help_text=_("""
        Attribute name on the parent table's base model.

        If this column is sortable, that attribute is used to sort this
        column.

        In case no value is given in "Template" field,
        this attribute will be used to get information from model
        instances.

        Warning, if you want to make this column sortable, you need to
        provide this value.

        This value can contain dot notation to reference fields in related
        models.
        """),
        blank=True, null=True)

    template = models.TextField(
        verbose_name=_("Template"),
        default="{{ value }}",
        null=True,
        blank=True,
        help_text=_("""If empty, the value of the object's attribute from
        "Attribute name" field will be used instead.

        Template will get following values in it's context:
        - *record*  -- data record for the current row
        - *value*   -- value from `record` that corresponds to the current column
        - *default* -- appropriate default value to use as fallback
        """))

    attrs = JSONField(
        verbose_name=_("HTML attributes"),
        blank=True,
        null=True
    )

    display_totals = models.BooleanField(
        default=False,
        verbose_name=_("Display totals"),
        help_text=_("Display column totals in footer. For columns without "
                    "Attribute name, this will be total number of the rows.")
    )

    strip_html_on_export = models.BooleanField(
        default=True,
        verbose_name=_("Strip HTML on export"),
        help_text=_("""Strip HTML tags when exporting to other, non-browser
        formats, like MS Word or MS Excel. """)
    )

    exclude_from_export = models.BooleanField(
        default=False,
        verbose_name=_("Exclude from export"),
        help_text=_("Exclude this column when exporting to other, non-browser"
                    "formats, like MS Word or MS Excel")
    )

    footer_template = models.TextField(
        verbose_name=_("Footer template"),
        default="{{ value }}",
        blank=True,
        null=True,
        help_text=_("""
        Template for footer. Used only if "Display totals" is enabled. It is
        rendered with 3 variables:
        - *count* -- total count of rows in the table,
        - *value* -- sum of this column's values (or row count if non-numeric),
        - *error* -- string representation of exception in case an exception
        occurs during addition of column's values.

        So, if the column values are numbers, just use {{ value }}. If you want
        to output number of rows, just use {{ count }}. """)
    )

    class Meta:
        unique_together = ('parent', 'id', 'position')
        ordering = ('position',)
        verbose_name = _("Column")
        verbose_name_plural = _("Columns")

    def clean(self):
        parent_model = self.parent.base_model.model_class()

        if self.attr_name:

            path = self.attr_name.split(".")

            current_model = parent_model

            for attr_name in path:
                try:
                    current_model = getattr(current_model, attr_name)

                    if hasattr(current_model, 'get_queryset'):
                        current_model = current_model.get_queryset().model

                except Exception as e:
                    raise ValidationError(
                        {"attr_name": [
                            ValidationError(
                                _(
                                    "'%(attr_name)s' is not a valid value for base "
                                    "model '%(base_model)s' (exception: %("
                                    "exception)s). "),
                                params={
                                    "attr_name": self.attr_name,
                                    "base_model": parent_model,
                                    "exception": e
                                }
                            )]})

        if self.sortable and not self.attr_name:
            raise ValidationError(
                {"attr_name": [
                    ValidationError(
                        _("You marked this column as sortable. You must "
                          "enter value into \"Attribute name\" field. ")
                    )
                ]}
            )

        if not self.attr_name and not self.template:
            raise ValidationError(
                {"attr_name": [
                    ValidationError(
                        _("You must either enter a template for this "
                          "column or an attribute name. "))],
                    "template": [
                        ValidationError(
                            _("You must either enter a template for this "
                              "column or an attribute name. "))
                    ]
                }
            )  # noqa

        if self.display_totals and not self.footer_template:
            raise ValidationError(
                {'footer_template': [
                    ValidationError(
                        _("If 'Display totals' is enabled, you should "
                          "provide a footer template. Perhaps try with a "
                          "very basic and default one, like '{{ value }}'. ")
                    )
                ]}
            )
